#define VERSION "14.2.95/V1.2 M.Mummert"
/*
 *	$Id: rateconv.c,v 1.2 1995/02/16 04:18:02 mummert Exp mummert $
 *
 *	RATECONV.C
 *
 *	Converting soundfiles/streams sampled at different rates
 *
 *	Use gcc -O to compile for maximum speed.
 *
 *
 *	Copyright (c) 1992, 1995 by Markus Mummert
 *
 *	Redistribution and use of this software, modifcation and inclusion
 *	into other forms of software are permitted provided that the following
 *	conditions are met:
 *
 *	1. Redistributions of this software must retain the above copyright
 *	   notice, this list of conditions and the following disclaimer.
 *	2. If this software is redistributed in a modified condition
 *	   it must reveal clearly that it has been modified.
 *	
 *	THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS''
 *	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 *	TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 *	PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *	PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *	PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 *	OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 *	USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 *	DAMAGE.
 *
 *
 *	history: 2.9.92		begin coding
 *		 5.9.92		fully operational
 *		 14.2.95 	provide BIG_ENDIAN, SWAPPED_BYTES_DEFAULT
 *				switches, Copyright note and References
 *
 *
 *	IMPLEMENTATION NOTES
 *
 *	Converting is done by interpolating the input samples in order to
 *      evaluate the represented continuous input function at sample 
 *	instances of the new rate (resampling). This is accomplished by a 
 *	FIR-filtering process. The rate conversion factor is determined by a
 *	rational factor, whose nominator and denominator are integers of
 *	almost arbitrary value, limited only by coefficient memory size.
 *
 *	General rate conversion:
 *
 *	    out(n*Tout) = SUM in(m*Tin) * g((n*d/u-m)*Tin) * Tin
 *		      over all m
 *
 *	FIR-based rate conversion formula for stream processing:
 *
 *			  L-1
 *	    out(n*Tout) = SUM in(A(i,n)*Tin) * g(B(i,n)*Tin) * Tin
 *			  i=0
 *
 *	    A(i,n) = i - [L/2] + [n*d/u]              
 *	           = i - [L/2] + [(n%u)*d/u] + [n/u]*d 
 *	    B(i,n) = n*d/u - [n*d/u] + [L/2] - i
 *	           =  ((n%u)*d/u)%1  + [L/2] - i
 *	    Tout   = Tin * d/u
 *
 *	  where:
 *	    n,i		running intergers
 *	    out(t)	output signal sampled at t=n*Tout
 *	    in(t)	input signal sampled at distances of Tin
 *	    u,d		up- and downsampling factor, integers
 *	    g(t)	interpolating function
 *	    L		FIR-length of realized g(t), integer
 *	    /		float-division-operator
 *	    %		float-modulo-operator
 *	    []		integer-operator
 *
 *	  note:
 *	    [L/2]	in A(i,n) can be omitted as a pure time shift
 *	    n%u		is a cyclic modulo-u counter clocked by out-rate
 *	    [n/u]*d	is a d-increment counter, advanced when n%u resets
 *	    B(i,n)*Tin	can take on L*u differnt values, at which g(t)
 *			has to be sampled as a coefficient array
 *
 *	Interpolation function design:
 *
 * 	    The interpolation function design is based on a sinc-function
 *	    multiplied by a gaussian function. The former determines the
 *	    cutoff frequency, the latter limits the necessary FIR-length by
 *	    pushing the outer skirts of the resulting impulse response below
 *	    quantizing error fast enough. The drawback is a smoothed resulting
 *	    cutoff inducing aliasing.
 *
 *	    g(t) = 2*fgk*sinc(pi*2*fgk*t) * exp(-pi*(2*fgg*t)**2)
 *
 *	  where:
 *	    fgk		cutoff frequency of sinc function in f-domain
 *	    fgg		key frequency of gaussian window in f-domain
 *			as the 6.82dB-down point
 *
 * 	  note:	    
 *	    Taking fin=1/Tin as the input sampling frequncy, it turns out
 *	    that only the ratios fgk/(fin/2) and fgg/(fin/2) are being needed
 *	    as parameters in the whole process, making no need for Tin,
 *	    Tout, fgk, fgg seperatly, as u and d determine rate conversion.
 *
 *	Numerical design:
 *
 *	    Samples are expected to be 16bit-signed integers in selectable
 *	    byte order. FIR-filtering is done on 32bit-signed integer
 *	    arithmetic. Coefficients are scaled to find new sample in the
 *	    high word of akkumulated FIR-sum.
 *
 *	    Interpolation can lead to sample magnitudes exceeding the input
 *	    maximum. Worst case is a step function with maximum level.
 *	    Here the ideal interpolation for resampling by means of the
 *	    sinc function has an overshoot of 9 percent. Sample overflow
 *	    can thus be avoided by a gain of about 0.9.
 *
 *
 * 	REFERENCES
 *
 *	    Crochiere, R. E., Rabiner, L. R.: "Multirate Digital Signal
 *	    Processing", Prentice-Hall, Englewood Cliffs, New Jersey, 1983
 *
 *	    Zwicker, E., Fastl, H.: "Psychoacoustics - Facts and Models",
 *	    Springer-Verlag, Berlin, Heidelberg, New-York, Tokyo, 1990
 */

#include <math.h>
#include <stdio.h>
#include <fcntl.h>

/*
 *	adaptable defines and globals
 */
#define	LONG		int
#define	WORD		short
#define	BYTE		char

#ifndef MAXUP
#define	MAXUP		0x400		/* MAXUP*MAXLENGTH worst case malloc */
#endif

#ifndef MAXLENGTH
#define	MAXLENGTH	0x400		/* max FIR length */
#endif

#define OUTBUFFSIZE 	MAXLENGTH
#define INBUFFSIZE	(2*MAXLENGTH)
#define sqr(a)		((a)*(a))

#ifndef BIG_ENDIAN			/* from Makefile */
#define LITTLE_ENDIAN
#endif

#ifdef SWAPPED_BYTES_DEFAULT		/* form Makefile */
int	g_swapflag = -1;		/* default byteswap else zero */
#else
int	g_swapflag = 0;
#endif

double	g_ampli = 0.9; 			/* default amplification factor */

/*
 *	globals
 */
int
	g_infilehandle = 0,		/* stdin */
	g_outfilehandle = 1,		/* stdout */
	g_firlen,			/* FIR-length */
	g_firdeg,			/* FIR-degree = FIR-length -1 */
	g_up,				/* upsampling factor */
	g_down				/* downsampling factor */
;

LONG
	g_sin[INBUFFSIZE],		/* input buffer */
	g_sout[OUTBUFFSIZE],		/* output buffer */
	*g_coep				/* coefficient array pointer */
;
double
	g_fsi,				/* input sampling frequency */
	g_fgk,				/* sinc-filter cutoff frequency */
	g_fgg				/* gaussian window key frequency */
;					/* (6.8dB down freq. in f-domain) */
	

/*
 *	evaluate sinc(x) = sin(x)/x safely
 */
double sinc(x)
double x;
{
	return(fabs(x) < 1E-50 ? 1.0 : sin(fmod(x,2*M_PI))/x);
}

/*
 *	evaluate interpolation function g(t) at t
 */
double interpol_func(t, fgk, fgg)
double t, fgk, fgg;
{
	return (2*fgk*sinc(M_PI*2*fgk*t)*exp(-M_PI*sqr(2*fgg*t)));
}

/*
 *	evaluate coefficient from i, q=n%u by sampling interpolation funtion 
 *	and scaling it for integer multiplication during FIR-filtering
 */
LONG coefficient(i, q, firlen, fgk, fgg, fsi, up, down, amp)
int i, firlen, q, up, down;
double fgk, fgg, fsi, amp;
{
	return(
	    (int)(0x10000 * amp *
		interpol_func(
		    (fmod(q*down/(double)up,1.) + floor(firlen/2.) - i) / fsi,
		    fgk,
		    fgg
		) / fsi
	    )
	);
}

/*
 *	I/O error handler, jumps to exit
 */
void ioerr_exit(p)
char *p;
{
	perror(p);
	free(g_coep);
	exit(-1);
}

/*
 *	transfer n LONGs from  s to d
 */
void transfer_int(s, d, n)
LONG *s, *d, n;
{
	LONG *e;

	if (n < 1)
		return;	
	e = d + n;
	while (d != e)
		*d++ = *s++;
}

/*
 *	zerofill n LONGs from s 
 */
void zerofill(s, n)
LONG *s;
{
	LONG *e;

	if (n < 1)
		return;	
	e = s + n;
	while (s != e)
		*s++ = 0;
}

/*
 *	convert buffer of n samples to LONGs
 */
sample_to_int(buff, n)
WORD *buff;
int n;
{
	WORD *s, *e;
	LONG *d;

	if (n < 1)
		return;	
	s = buff + n;
	d = (int*)buff + n;
	e = buff;
	while (s != e) {
		*--d = (int)(*--s); 
	}
}

/*
 *	convert buffer of n LONGs to samples
 */
void int_to_sample(buff, n)
WORD *buff;
int n;
{
	WORD *s, *e, *d;

	if (n < 1)
		return;	
	s = buff;
	d = buff;
	e = buff + n*2;
	while (s != e) {

#ifndef BIG_ENDIAN
		s++;
		*d++ = *s++; 
#else
		*d++ = *s++; 
		s++;
#endif
	}
}

/*
 *	swap bytes in buffer of n samples
 */
void byteswap_samples(buff, n)
BYTE *buff;
int n;
{
	BYTE *s, *d, *e;
	BYTE b;

	if (n < 1)
		return;	
	s = buff;
	d = buff;
	e = buff + n*2;
	while (s != e) {
		b = *s++;
		*d++ = *s++;
		*d++ = b;
	}
}

/*
 *	FIR-routine
 *	use gcc -O to compile for maximum speed
 */
LONG fir(inp, coep, firlen)
register LONG *inp, *coep;
int firlen;
{
	register LONG *endp, akku = 0;
	int n1 = (firlen / 8) * 8, n0 = firlen % 8;

	endp = coep + n1;
	while (coep != endp) {
		akku += *inp++ * *coep++;
		akku += *inp++ * *coep++;
		akku += *inp++ * *coep++;
		akku += *inp++ * *coep++;
		akku += *inp++ * *coep++;
		akku += *inp++ * *coep++;
		akku += *inp++ * *coep++;
		akku += *inp++ * *coep++;
	}

	endp = coep + n0;
	while (coep != endp) {
		akku += *inp++ * *coep++;
	}

	return(akku);
}

/*
 * 	filtering from input buffer to output buffer;
 *	returns number of processed samples in output buffer:
 *	if it is not equal to output buffer size,
 *	the input buffer is expected to be refilled upon entry, so that
 *	the last firlen numbers of the old input buffer are
 *	the first firlen numbers of the new input buffer;
 *	if it is equal to output buffer size, the output buffer
 *	is full and is expected to be stowed away;
 *
 */
int filtering_on_buffers(inp, insize, outp, outsize, coep, firlen, up, down)
LONG *inp, *outp, *coep;
int insize, outsize, firlen, up, down;
{
	static int inbaseidx = 0, inoffset = 0, cycctr = 0, outidx = 0;

	while (-1) {
		inoffset = (cycctr * down)/up;
		if ((inbaseidx + inoffset + firlen) > insize) {
			inbaseidx -= insize - firlen + 1;
			return(outidx);
		}
		outp[outidx++] =
			fir(inp + inoffset + inbaseidx,
				 coep + cycctr * firlen, firlen);
		cycctr++;
		if (!(cycctr %= up))
			inbaseidx += down;
		if (!(outidx %= outsize))
			return(outsize);
	}
}

/*
 *	intread, read and convert input sample format to integer
 */
int intread(hd, buff, n)
int hd, n;
LONG *buff;
{
	int rd;

	if ((rd = read(g_infilehandle, buff, n*sizeof(WORD))) <= 0)
		return(rd);
	if (g_swapflag)
		byteswap_samples(buff, n);
	sample_to_int(buff,n);
	return(rd/sizeof(WORD));
}

/*
 *	intwrite, do some conversion jobs and write
 */
int intwrite(hd, buff, n)
int hd, n;
LONG *buff;
{
	int written;

	int_to_sample(buff, n);
	if (g_swapflag)
		byteswap_samples(buff, n);
	if ((written = write(g_outfilehandle, buff, n*sizeof(WORD))) <= 0)
		return(written);
	return(written/sizeof(WORD));
}

/*
 *	print usage
 */
void print_usage(argc, argv)
int argc;
char *argv[];
{
	fprintf(stderr, "\n");
	fprintf(stderr, "    sample rate conversion from stdin to ");
	fprintf(stderr, "stdout %s\n",VERSION);
	fprintf(stderr, "    usage: %s [-h|-l] <fsin> <fgk> <fgg> ", argv[0]);
	fprintf(stderr, "<lenght> <up> <down> [<gain>]\n");
	fprintf(stderr, "      -h -l     sample format HB,LB ");
	if (g_swapflag)
		fprintf(stderr, "(default) or LB,HB\n");
	else
		fprintf(stderr, "or LB,HB (default)\n");
	fprintf(stderr, "      <fsin>    input sampling frequency in Hz\n");
	fprintf(stderr, "      <fgk>     sinc-filter cutoff frequency\n");
	fprintf(stderr, "      <fgg>     gaussian-window key frequency ");
	fprintf(stderr, "(6.8dB-down point)\n");
	fprintf(stderr, "      <length>  lenght of IR of resulting FIR-");
	fprintf(stderr, "filter (1...%d)\n",MAXLENGTH);
	fprintf(stderr, "      <up>      upsampling factor ");
	fprintf(stderr, "(1...%d)\n",MAXUP);
	fprintf(stderr, "      <down>    downsampling factor\n");
	fprintf(stderr, "      <gain>    gain adjust (default ");
	fprintf(stderr, "%4.2lf safe on filter overshoot)\n", g_ampli);
	fprintf(stderr, "\n");
}

/*
 *	check command line parameters and get their values
 */
int check_args(argc, argv)
int argc;
char *argv[];
{
	int off = 0;

	if (argc < 7 || argc > 9) 
		return(-1);
	if (argv[1][0] == '-') {
		if ((argv[1][1] != 'h' && argv[1][1] != 'l') ||
		    argv[1][2] != 0)
			return(-1);
		off = 1;
		if (argv[1][1] == 'h')
			g_swapflag = -1;
		if (argv[1][1] == 'l')
			g_swapflag = 0;
	}
	sscanf(argv[1+off],"%lf",&g_fsi);
	sscanf(argv[2+off],"%lf",&g_fgk);
	sscanf(argv[3+off],"%lf",&g_fgg);
	sscanf(argv[4+off],"%d",&g_firlen);
	if (g_firlen < 1 || g_firlen > MAXLENGTH)
		return(-1);
	sscanf(argv[5+off],"%d",&g_up);
	if (g_up < 1 || g_up > MAXUP)
		return(-1);
	sscanf(argv[6+off],"%d",&g_down);
	if (argc > 7+off)
		sscanf(argv[7+off],"%lf",&g_ampli);
	return(0);
}

/*
 *	make coefficient array
 */
void make_coe()
{
	int i, q;

	for (i = 0; i < g_firlen; i++) {
	    for (q = 0; q < g_up; q++) {
		g_coep[q * g_firlen + i] = coefficient(i, q, g_firlen,
		    g_fgk, g_fgg, g_fsi, g_up, g_down, g_ampli);
	    }
	}
}

/*
 *	main
 */
main(argc, argv)
int argc;
char *argv[];
{
	extern char *malloc();
	int insize = 0, outsize = 0;

	if (check_args(argc, argv)) {
	    print_usage(argc, argv);
	    exit(-1);
	}
	g_firdeg = g_firlen - 1;
	if ((g_coep = (int*)malloc(g_firlen * g_up * sizeof(int))) == NULL) {
		fprintf(stderr, "cannot allocate coefficient memory\n");
		exit(-1);
	}
	make_coe();
	zerofill(g_sin, g_firdeg);
	do {
	    insize = intread(g_infilehandle, g_sin + g_firdeg,
		INBUFFSIZE - g_firdeg);
	    if (insize < 0 || insize > INBUFFSIZE - g_firdeg) 
		ioerr_exit("stdin");
	    do {
		outsize = filtering_on_buffers(g_sin, g_firdeg + insize,
		    g_sout, OUTBUFFSIZE, g_coep, g_firlen, g_up, g_down);
		if (outsize != OUTBUFFSIZE) {
	    	    transfer_int(g_sin + insize, g_sin, g_firdeg);
		    break;
		}
		if (intwrite(g_outfilehandle, g_sout, outsize) != outsize) 
		    ioerr_exit("stdout");
	    } while (-1);
 	} while (insize > 0);
	zerofill(g_sin + g_firdeg, g_firdeg);
	do {
	    outsize = filtering_on_buffers(g_sin, g_firdeg + g_firdeg,
		g_sout, OUTBUFFSIZE, g_coep, g_firlen, g_up, g_down);
	    if (intwrite(g_outfilehandle, g_sout, outsize) != outsize) 
		ioerr_exit("stdout");
	} while (outsize == OUTBUFFSIZE); 

	if (close(g_infilehandle)) 
		ioerr_exit("stdin");
	if (close(g_outfilehandle))
		ioerr_exit("stdout");

	free(g_coep);
	exit(0);
}
/*
 *	EOT
 */
